BROKER_URL = 'redis://'
#ELERY_ENABLE_UTC = True
#CELERY_TIMEZONE = 'UTC'

CELERYBEAT_SCHEDULER = 'redbeat.RedBeatScheduler'
CELERYBEAT_MAX_LOOP_INTERVAL = 5  # redbeat likes fast loops

#REDBEAT_REDIS_URL = BROKER_URL
#REDBEAT_KEY_PREFIX = 'redbeat'
#REDBEAT_SCHEDULE_KEY = REDBEAT_KEY_PREFIX + ':schedule'
#REDBEAT_STATICS_KEY = REDBEAT_KEY_PREFIX + ':statics'
#REDBEAT_LOCK_KEY = REDBEAT_KEY_PREFIX + ':lock'
#REDBEAT_LOCK_TIMEOUT = CELERYBEAT_MAX_LOOP_INTERVAL * 5
